import {Ref, ref, watch} from "vue";
import useExpr1Store from "@/store/expr1/expr1Store.ts";
import {storeToRefs} from "pinia";
import {Operation} from "@/store/expr1/defaultOperation.ts";

export enum Selector {
    FirstFit,
    BestFit,
    WorstFit
}

export class Ram {
    _memSize: Ref<number, number>;
    _memory: Ref<Block[], Block[]>;
    _operations: Ref<Operation[], Operation[]>;
    _selectors: ((operation: Operation) => number)[];
    _currentSelector: number = 0;
    _nextOperation: number = 0;
    _currentState: Ref<number, number>;
    _allStates: Ref<Block[][], Block[][]> = ref([]);
    _releaseHistory: Map<number, number>

    static OperationToProcess(operation: Operation): Block {
        return new Block(operation.id, operation.memorySize);
    }

    constructor() {
        let store = useExpr1Store();
        let {operations, memSize, tableChanged} = storeToRefs(store);
        this._operations = operations;
        this._memSize = memSize;
        this._memory = ref([new Block(0, memSize.value, "Space")]);
        this._selectors = [this.FirstFit.bind(this), this.BestFit.bind(this), this.WorstFit.bind(this)];
        this._currentState = storeToRefs(store).currentState;
        this._allStates.value.push(this._memory.value.map(x => x));
        this._releaseHistory = new Map();
        let {failInfo} = storeToRefs(store);
        while (this._nextOperation < operations.value.length) {
            this.Process();
            this._allStates.value.push(this._memory.value.map(x => Block.Clone(x)));
        }
        watch(tableChanged, (val) => {
            if (!val)
                return;
            this._nextOperation = 0;
            this._allStates.value = [];
            failInfo.value = "";
            this._releaseHistory = new Map();
            this._memory.value = [new Block(0, memSize.value, "Space")];
            this._allStates.value.push(this._memory.value.map(x => Block.Clone(x)));
            while (this._nextOperation < this._operations.value.length) {
                try {
                    this.Process();
                    this._allStates.value.push(this._memory.value.map(x => Block.Clone(x)));
                } catch (e:any) {
                    e = e as Error;
                    console.log(this._nextOperation, e);
                    failInfo.value = "任务" + this._nextOperation + '\t' +e.message;
                    while (this._nextOperation < operations.value.length) {
                        this._allStates.value.push(this._memory.value.map(x => Block.Clone(x)));
                        this._nextOperation++;
                    }
                    break;
                }
            }
            tableChanged.value = false;
        });
    }

    set CurrentSelector(value: Selector) {
        this._currentSelector = value;
        console.log(this._currentSelector);
    }

    GetMemory() {
        if (this._currentState.value >= this._allStates.value.length)
            this._currentState.value = this._allStates.value.length - 1;
        return this._allStates.value[this._currentState.value];
    }

    GetReleaseHistory(processID: number) {
        if (this._releaseHistory.has(processID))
            return this._releaseHistory.get(processID)!;
        else
            return '-';
    }

    FirstFit(operation: Operation): number {
        for (let i = 0; i < this._memory.value.length; i++) {
            if (this._memory.value[i].Type === "Process") {
                continue;
            }
            if (this._memory.value[i].MemorySize >= operation.memorySize) {
                return i;
            }
        }
        return -1;
    }

    BestFit(operation: Operation): number {
        let memories: Map<number, number> = new Map();
        for (let i = 0; i < this._memory.value.length; i++) {
            if (this._memory.value[i].Type === "Process") {
                continue;
            }
            memories.set(i, this._memory.value[i].MemorySize);
        }
        let sorted = Array.from(memories).sort((a, b) => a[1] - b[1]);
        for (let i = 0; i < sorted.length; i++) {
            if (sorted[i][1] >= operation.memorySize) {
                return sorted[i][0];
            }
        }
        return -1;
    }

    WorstFit(operation: Operation): number {
        let memories: Map<number, number> = new Map();
        for (let i = 0; i < this._memory.value.length; i++) {
            if (this._memory.value[i].Type === "Process") {
                continue;
            }
            memories.set(i, this._memory.value[i].MemorySize);
        }
        let sorted = Array.from(memories).sort((a, b) => b[1] - a[1]);
        for (let i = 0; i < sorted.length; i++) {
            if (sorted[i][1] >= operation.memorySize) {
                return sorted[i][0];
            }
        }
        return -1;
    }

    Process() {
        if (this._nextOperation >= this._operations.value.length)
            return;
        let operation = this._operations.value[this._nextOperation];
        if (operation.type === 'request' && operation.memorySize === 0) {
            this._nextOperation++;
            return;
        }
        this._nextOperation++;
        let process = Ram.OperationToProcess(operation);
        if (operation.type === 'release') {
            let index = this._memory.value.findIndex((value) => value.Type === "Process" && value.ProcessID == process.ProcessID);
            if (index == -1) {
                throw Error("Process not found");
            }
            if (process.MemorySize !== this._memory.value[index].MemorySize) {
                console.warn("Memory size not match");
                process.MemorySize = this._memory.value[index].MemorySize;
            }
            this._releaseHistory.set(process.ProcessID, this._memory.value[index].MemorySize);
            let space = new Block(this._memory.value[index].ProcessID, process.MemorySize, "Space");
            space.Start = this._memory.value[index].Start;
            this._memory.value[index] = space;
            if (index > 0 && this._memory.value[index - 1].Type === "Space") {
                let prev = this._memory.value[index - 1];
                prev.MemorySize += space.MemorySize;
                this._memory.value.splice(index, 1);
                index--;
                space = prev;
            }
            if (index < this._memory.value.length - 1 && this._memory.value[index + 1].Type === "Space") {
                let next = this._memory.value[index + 1];
                space.MemorySize += next.MemorySize;
                this._memory.value.splice(index + 1, 1);
            }
            return;
        }

        let index = this._selectors[this._currentSelector](operation);
        if (index == -1) {
            throw Error("No enough memory");
        }
        let space = this._memory.value[index];
        process.Start = space.Start;
        if (space.MemorySize == process.MemorySize) {
            this._memory.value.splice(index, 1);
        } else {
            space.Start += process.MemorySize;
            space.MemorySize -= process.MemorySize;
        }
        this._memory.value.push(process);
        this._memory.value.sort((a, b) => a.Start - b.Start);
    }
}

export class Block {
    ProcessID = 0;
    MemorySize = 0;
    Start = 0;
    Type: "Process" | "Space" = "Process";

    constructor(processID: number, memorySize: number, type: "Process" | "Space" = "Process") {
        this.ProcessID = processID;
        this.MemorySize = memorySize;
        this.Type = type;
    }

    static Clone(block: Block) {
        let newBlock = new Block(block.ProcessID, block.MemorySize, block.Type);
        newBlock.Start = block.Start;
        return newBlock;
    }
}
